"""
Main application module
"""
import argparse
import logging
import logging.handlers
import os
import tempfile

from launchpadlib.launchpad import Launchpad, \
        EDGE_SERVICE_ROOT, STAGING_SERVICE_ROOT
from gi.repository import GObject, Gtk, Gdk
from sqlalchemy.sql import exists

#Use threads
GObject.threads_init()
Gdk.threads_init()

from .uifile import UIFile
from .dialog import ProjectDialog, ProgressDialog, DbUpdateDialog
from .db import Database
from .db.model import Project, Milestone, Person, \
    Bug, BugTask, BugTaskView, Tag, BugViewTag
from .db.update import ProjectUpdater
from .notebook import Notebook
from .notebook.tab import HomeTab
from .tree.model import TreeStore
from .tree.model import TreeStoreColumnMetadata as ColumnMetadata
from .util import critical_section, Enumeration


class Application(UIFile):
    """
    Application implementation
    """
    DEFAULT_WINDOW_TITLE = 'Bug Time Tracker'

    def __init__(self):
        super(Application, self).__init__('bugtimetracker.ui')

        self._init_stores()
        self._init_notebook()

        self.db = None
        self.window.show_all()

    def _init_notebook(self):
        """
        Initialize right notebook that will display launchpad information
        """
        right_notebook = Notebook(self)
        self.paned.add2(right_notebook)
        self.right_notebook = right_notebook

    def _init_stores(self):
        """
        Initialize stores that will contain launchpad information
        """
        projects_store = TreeStore('projects',
                                   (ColumnMetadata('id', int, key=True),
                                    ColumnMetadata('name', str, required=True),
                                   ))
        projects_store.connect('row-inserted', self.project_row_inserted_cb)
        projects_store.connect('row-deleted', self.project_row_deleted_cb)
        self.projects_treeview.set_model(projects_store)
        self.projects_store = projects_store

        people_store = TreeStore('people',
                                 (ColumnMetadata('id', int, key=True),
                                  ColumnMetadata('name', str, required=True),
                                  ColumnMetadata('display_name', str),
                                  ColumnMetadata('is_team', bool),
                                 ))
        self.people_treeview.set_model(people_store)
        self.people_store = people_store

        bugs_store = TreeStore('bugs',
                               (ColumnMetadata('id', int, key=True),
                                ColumnMetadata('title', str)))
        self.bugs_treeview.set_model(bugs_store)
        self.bugs_store = bugs_store

    def main(self):
        args = self._parse_arguments()
        self.launchpad = Launchpad.login_with('bugtimetracker', args.service)
        self._configure_logging(args.log_level)
        Gtk.main()

    def _parse_arguments(self):
        """
        Parse arguments from the command line
        """
        description = ('Get bug information from launchpad '
                'and display it nicely')
        parser = argparse.ArgumentParser(description=description)
        log_levels = ['debug', 'info', 'warning', 'error', 'critical']
        parser.add_argument('--log-level', dest='log_level_str',
                default='info',
                choices=log_levels,
                help='Log level (%(default)s by default)')
        services = {'edge': EDGE_SERVICE_ROOT,
                'staging': STAGING_SERVICE_ROOT}
        parser.add_argument('--service', dest='service_str',
                default='edge',
                choices=services.keys(),
                help='Select launchpad service to use in queries')
        args = parser.parse_args()
        args.log_level = getattr(logging, args.log_level_str.upper())
        args.service = services[args.service_str]
        return args

    def _configure_logging(self, log_level):
        """
        Configure logging to use sys.stderr
        and a file in the default temporary directory
        """
        # Configure logging
        logger = logging.getLogger()
        logger.handlers = []
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)s: %(message)s')

        # Log to sys.stderr using log level passed through command line
        log_handler = logging.StreamHandler()
        log_handler.setFormatter(formatter)
        log_handler.setLevel(log_level)
        logger.addHandler(log_handler)

        # Log to file using DEBUG log level
        log_filename = os.path.join(tempfile.gettempdir(),
                'bugtimetracker.log')
        log_filename_exists = os.path.exists(log_filename)
        log_handler = (logging.handlers
                       .RotatingFileHandler(log_filename, mode='a+',
                                            backupCount=3))
        formatter = logging.Formatter('%(asctime)s - %(levelname)s: '
                                      '%(message)s')
        log_handler.setFormatter(formatter)
        log_handler.setLevel(logging.DEBUG)
        logger.addHandler(log_handler)

        if log_filename_exists:
            log_handler.doRollover()

    def _get_filename(self, action):
        """
        Get database filename
        """
        if action == Gtk.FileChooserAction.OPEN:
            dialog_buttons = (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                    Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        elif action == Gtk.FileChooserAction.SAVE:
            dialog_buttons = (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                    Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
        dialog = Gtk.FileChooserDialog('Select database file...',
                parent=self.window,
                action=action,
                buttons=dialog_buttons)
        dialog.set_do_overwrite_confirmation(True)

        for name, pattern in (('Database files (*.sqlite)', '*.sqlite'),
                ('All files', '*')):
            file_filter = Gtk.FileFilter()
            file_filter.set_name(name)
            file_filter.add_pattern(pattern)
            dialog.add_filter(file_filter)

        response_id = dialog.run()
        if response_id == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
        else:
            filename = None
        dialog.destroy()

        return filename

    def _populate_stores(self):
        """
        Populate treeview stores according to database contents
        """
        def target(progressbar):
            prefixes = ('projects', 'people', 'bugs')
            stores = [getattr(self, '{0}_store'.format(prefix))
                      for prefix in prefixes]
            treeviews = [getattr(self, '{0}_treeview'.format(prefix))
                         for prefix in prefixes]

            with critical_section():
                for store in stores:
                    store.clear()

                for treeview in treeviews:
                    treeview.freeze_child_notify()
                    treeview.set_model(None)

            session = self.db.session
            project_count = session.query(Project).count()
            person_count = session.query(Person).count()
            bug_count = session.query(Bug).count()
            total_count = project_count + person_count + bug_count
            row_enumeration = Enumeration()

            projects = session.query(Project).order_by(Project.name)
            for index, project in row_enumeration(projects):
                with critical_section():
                    project_row = (project.id, project.name)
                    project_iter = self.projects_store.append(None,
                                                              project_row)
                    milestones = (session.query(Milestone)
                                  .filter_by(project=project)
                                  .order_by(Milestone.name))
                    for milestone in milestones:
                        milestone_row = (milestone.id, milestone.name)
                        self.projects_store.append(project_iter, milestone_row)

                    progress = float(index + 1) / total_count
                    progressbar.set_fraction(progress)

            people = session.query(Person).order_by(Person.name)
            for index, person in row_enumeration(people):
                with critical_section():
                    self.people_store.append(None,
                            (person.id, person.name, person.display_name,
                             person.is_team))

                    progress = float(index + 1) / total_count
                    progressbar.set_fraction(progress)

            bugs = session.query(Bug).order_by(Bug.id)
            for index, bug in row_enumeration(bugs):
                with critical_section():
                    self.bugs_store.append(None, (bug.id, bug.title,))

                    progress = float(index + 1) / total_count
                    progressbar.set_fraction(progress)

            with critical_section():
                for treeview, store in zip(treeviews, stores):
                    treeview.set_model(store)
                    treeview.thaw_child_notify()

        dialog = ProgressDialog(self.window, 'Widget update')
        dialog.run(target)

    def new_cb(self, widget):
        """
        Create new database file
        """
        self.projects_store.clear()
        self.people_store.clear()
        self.bugs_store.clear()
        self.right_notebook.close_all_tabs()
        self.db = Database()

        self.save_toolbutton.set_sensitive(False)
        self.save_menuitem.set_sensitive(False)
        self.save_as_toolbutton.set_sensitive(True)
        self.save_as_menuitem.set_sensitive(True)
        self.close_toolbutton.set_sensitive(True)
        self.close_menuitem.set_sensitive(True)
        self.home_toolbutton.set_sensitive(True)
        self.add_project_toolbutton.set_sensitive(True)

        self.window.set_title('New Database - {0}'
                              .format(self.DEFAULT_WINDOW_TITLE))

    def open_cb(self, widget):
        """
        Open database file
        """
        filename = self._get_filename(Gtk.FileChooserAction.OPEN)
        if filename:
            self.db = Database(filename)
            self.right_notebook.close_all_tabs()
            self._populate_stores()
            self.window.set_title('{0} - {1}'
                                  .format(filename, self.DEFAULT_WINDOW_TITLE))

            self.save_as_toolbutton.set_sensitive(True)
            self.save_as_menuitem.set_sensitive(True)
            self.close_toolbutton.set_sensitive(True)
            self.close_menuitem.set_sensitive(True)
            self.home_toolbutton.set_sensitive(True)
            self.add_project_toolbutton.set_sensitive(True)

    def close_cb(self, widget):
        """
        Close database file
        """
        self.db.close()
        self.db = None
        self.projects_store.clear()
        self.people_store.clear()
        self.bugs_store.clear()
        self.right_notebook.close_all_tabs()
        self.window.set_title(self.DEFAULT_WINDOW_TITLE)

        self.add_project_toolbutton.set_sensitive(False)
        self.save_toolbutton.set_sensitive(False)
        self.save_menuitem.set_sensitive(False)
        self.save_as_toolbutton.set_sensitive(False)
        self.save_as_menuitem.set_sensitive(False)
        self.close_toolbutton.set_sensitive(False)
        self.close_menuitem.set_sensitive(False)
        self.home_toolbutton.set_sensitive(False)

    def save_cb(self, widget):
        """
        Save updates to database file
        """
        self.db.save()
        self._saved()

    def save_as_cb(self, widget):
        """
        Save updates to a different database file
        """
        filename = self._get_filename(Gtk.FileChooserAction.SAVE)
        if filename:
            self.db.save_as(filename)
            self._saved()

    def _saved(self):
        """
        Update window title and disable save widgets
        """
        self.save_toolbutton.set_sensitive(False)
        self.save_menuitem.set_sensitive(False)

        # Unset modification mark to window title
        self.window.set_title('{0} - {1}'
                              .format(self.db.filename,
                                      self.DEFAULT_WINDOW_TITLE))

    def projects_selection_changed_cb(self, selection):
        """
        Update projects toolbar buttons sensitivity
        """
        _, selected = selection.get_selected()
        if selected is not None:
            row = self.projects_store[selected]
            # Look at depth to make sure that remove project button
            # isn't sensitive when a milestone is selected
            depth = row.path.get_depth()
            self.remove_project_toolbutton.set_sensitive(depth == 1)

    def project_row_inserted_cb(self, store, path, iterator):
        """
        Enable updates when at least one project is being tracked
        """
        self.update_toolbutton.set_sensitive(len(store) > 0)

    def project_row_deleted_cb(self, store, path):
        """
        Disable updates when at no project is being tracked
        """
        self.update_toolbutton.set_sensitive(len(store) > 0)

    def add_project_cb(self, widget):
        """
        Enable tracking a new launchpad project
        """
        dialog = ProjectDialog(self.window, self.launchpad, self.db.session)
        project_name = dialog.run()
        if project_name is None:
            return

        # Add project to database
        lp_project = self.launchpad.projects[project_name]
        project = Project(lp_project.name,
                          lp_project.date_created.replace(tzinfo=None))
        session = self.db.session
        session.add(project)
        session.commit()

        # Add project to model
        self.projects_store.append(None, (project.id, project_name,))

        self._modified()

    def remove_project_cb(self, widget):
        """
        Disable tracking a selected launchpad project
        """
        session = self.db.session
        notebook = self.right_notebook

        # Get currently selected project
        selection = self.projects_treeview_selection
        _, selected = selection.get_selected()
        project_row = self.projects_store[selected]

        # Delete project from database
        # (this will cascade delete milestones, bug tasks and bug task views)
        project = session.query(Project).get(project_row.id)
        milestones = project.milestone_set
        for milestone in milestones:
            notebook.find_and_close_tab(milestone)
        notebook.find_and_close_tab(project)
        project_row.remove()
        session.delete(project)

        # Remove bugs not referenced by any bug task
        # (this will cascade delete bug views and bug view tags)
        bugs = (session.query(Bug)
                .filter(~exists().where(Bug.id == BugTask.bug_id))
                .all())
        for bug in bugs:
            row_reference = self.bugs_store.index[bug.id]
            row_reference.row.remove()
            notebook.find_and_close_tab(bug)
            session.delete(bug)

        # Remove tags not referenced by any bug view tags)
        tags = (session.query(Tag)
                .filter(~exists()
                        .where(Tag.id == BugViewTag.tag_id)))
        for tag in tags:
            session.delete(tag)

        # Remove people not referenced by any bug task view
        people = (session.query(Person)
                  .filter(~exists()
                          .where(Person.id == BugTaskView.assignee_id))
                  .all())
        for person in people:
            row_reference = self.people_store.index[person.id]
            row_reference.row.remove()
            notebook.find_and_close_tab(person)
            session.delete(person)

        session.commit()
        self._modified()

    def update_cb(self, widget):
        """
        Update bug information
        """
        session = self.db.session
        project_updaters = [ProjectUpdater(self.launchpad, project)
                            for project in session.query(Project).all()]

        dialog = DbUpdateDialog(self.window)
        response_id, db_updated = dialog.run(project_updaters)
        if db_updated:
            if response_id == Gtk.ResponseType.OK:
                self.db.session.commit()
                logging.info('Update committed')
                self._populate_stores()
                self._modified()
            else:
                self.db.session.rollback()
                logging.info('Update rolled back')

    def _modified(self):
        """
        Update window title and enable save widgets if needed
        """
        self.right_notebook.refresh_all_tabs()

        # Enable save only if database file is opened
        if self.db.filename:
            self.save_toolbutton.set_sensitive(True)
            self.save_menuitem.set_sensitive(True)

        # Set modification mark to window title
        title = self.window.get_title()
        if not title.startswith('*'):
            self.window.set_title('*' + title)

    def home_cb(self, widget):
        """
        Open home tab
        """
        notebook = self.right_notebook
        notebook.open_tab(HomeTab.id, HomeTab, (self,))

    def quit_cb(self, menuitem):
        """
        Exit from application
        """
        logging.debug('Quit {0}'.format(self))
        self.window.destroy()

    def window_destroy_cb(self, window):
        """
        Quit from application
        """
        if self.db:
            self.db.close()
        logging.debug('Destroy {0}'.format(self))
        Gtk.main_quit()
