"""
Database module takes care of opening and saving data to a database file
"""
import logging
import tempfile
import os
import shutil
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from . import model

ALL_BUG_STATUSES = [
    'New',
    'Incomplete',
    'Opinion',
    'Invalid',
    "Won't Fix",
    'Confirmed',
    'Triaged',
    'In Progress',
    'Fix Committed',
    'Fix Released',
    'Expired',
    ]

ALL_BUG_IMPORTANCES = [
    'Critical',
    'High',
    'Medium',
    'Low',
    'Wishlist',
    'Undecided',
    ]


class Database(object):
    CONNECT_ARGS = {'check_same_thread': False}

    def __init__(self, filename=None):
        """
        Create new database or open an existing one
        """
        if filename is None:
            engine, session, temp_filename = self._create()
        else:
            engine, session, temp_filename = self._open(filename)
        self.filename = filename
        self.engine = engine
        self.session = session
        self.temp_filename = temp_filename

    def _create(self):
        """
        Create new empty database
        """
        _, temp_filename = tempfile.mkstemp(prefix='bugtimetracker-',
                                            suffix='.sqlite')
        logging.debug('Creating temporary database file at {0!r}...'
                      .format(temp_filename))
        engine = create_engine('sqlite:///{filename}'
                               .format(filename=temp_filename),
                               connect_args=self.CONNECT_ARGS)
        model.Base.metadata.create_all(engine)
        Session = sessionmaker(bind=engine)
        session = Session()

        for name in ALL_BUG_STATUSES:
            status = model.Status(name=name)
            session.add(status)

        for name in ALL_BUG_IMPORTANCES:
            importance = model.Importance(name=name)
            session.add(importance)

        session.commit()
        return engine, session, temp_filename

    def _open(self, filename):
        """
        Open an existing database
        """
        logging.debug('Opening database at {0!r}...'
                      .format(filename))

        # Create a temporary copy of the database
        # to prevent changes from bein saved on commit
        # (the user might not want to save changes
        # after adding/removing projects to the list)
        _, temp_filename = tempfile.mkstemp(prefix='bugtimetracker-',
                                            suffix='.sqlite')
        logging.debug('Creating temporary database file at {0!r}...'
                      .format(temp_filename))
        shutil.copyfile(filename, temp_filename)
        engine = create_engine('sqlite:///{filename}'
                               .format(filename=temp_filename),
                               connect_args=self.CONNECT_ARGS)
        Session = sessionmaker(bind=engine)
        session = Session()
        return engine, session, temp_filename

    def save(self):
        """
        Save changes to the opened database file
        """
        shutil.copyfile(self.temp_filename, self.filename)

    def save_as(self, filename):
        """
        Save changes to a given filename
        """
        self.filename = filename
        self.save()

    def close(self):
        """
        Delete database temporary file when is no longer needed
        """
        self.session.close()
        self.engine.dispose()
        if os.path.exists(self.temp_filename):
            logging.debug('Removing temporary database file at {0!r}...'
                          .format(self.temp_filename))
            os.remove(self.temp_filename)
