from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, ForeignKey, UniqueConstraint
from sqlalchemy import Boolean, String, Integer, DateTime
from sqlalchemy.orm import relationship, backref

Base = declarative_base()


class Project(Base):
    __tablename__ = 'project'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True, nullable=False)
    date_created = Column(DateTime)
    date_updated = Column(DateTime)

    def __init__(self, name, date_created):
        self.name = name
        self.date_created = date_created

    def __repr__(self):
        return '<Project: {0}>'.format(self.name)


class Series(Base):
    __tablename__ = 'series'
    __table_args__ = (UniqueConstraint('name', 'project_id'),)
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    project_id = Column(Integer,
                        ForeignKey('project.id'),
                        nullable=False)
    project = relationship(Project,
                           backref=backref('series_set',
                                           lazy='dynamic',
                                           cascade='all'))

    def __init__(self, name, project):
        self.name = name
        self.project = project

    def __repr__(self):
        return ('<Series: {series} ({project})>'
                .format(project=self.project.name,
                        series=self.name))


class Milestone(Base):
    __tablename__ = 'milestone'
    __table_args__ = (UniqueConstraint('name', 'project_id'),)
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    project_id = Column(Integer,
                        ForeignKey('project.id'),
                        nullable=False)
    project = relationship(Project,
                           backref=backref('milestone_set',
                                           lazy='dynamic',
                                           cascade='all'))

    def __init__(self, name, project):
        self.name = name
        self.project = project

    def __repr__(self):
        return ('<Milestone: {milestone} ({project})>'
                .format(project=self.project.name,
                        milestone=self.name))


class Status(Base):
    __tablename__ = 'status'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<Status: {0}>'.format(self.name)


class Importance(Base):
    __tablename__ = 'importance'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<Importance: {0}>'.format(self.name)


class Person(Base):
    __tablename__ = 'person'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    display_name = Column(String)
    is_team = Column(Boolean)

    def __init__(self, name, display_name=None, is_team=False):
        self.name = name
        self.display_name = display_name
        self.is_team = is_team

    def __repr__(self):
        return ('<Person: {0} ({1})>'
                .format(self.display_name,
                        self.name))


class Bug(Base):
    __tablename__ = 'bug'
    id = Column(Integer, primary_key=True, autoincrement=False)
    title = Column(String)

    def __init__(self, id, title):
        self.id = id
        self.title = title

    def __repr__(self):
        return '<Bug: {0}>'.format(self.id)


class Tag(Base):
    __tablename__ = 'tag'
    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True, nullable=False)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<Tag: {0}>'.format(self.name)


class BugView(Base):
    __tablename__ = 'bugview'
    __table_args__ = (UniqueConstraint('datechanged', 'bug_id'),)
    id = Column(Integer, primary_key=True)
    datechanged = Column(DateTime, nullable=False)
    bug_id = Column(Integer, ForeignKey('bug.id'))
    bug = relationship(Bug,
                       backref=backref('bugview_set',
                                       lazy='dynamic',
                                       order_by=datechanged.desc(),
                                       cascade='all'))

    def __init__(self, datechanged, bug):
        self.datechanged = datechanged
        self.bug = bug

    def __repr__(self):
        return ('<BugView: {0}@{1}>'
                .format(self.bug.id, self.datechanged))


class BugViewTag(Base):
    __tablename__ = 'bugviewtag'
    bugview_id = Column(Integer, ForeignKey('bugview.id'), primary_key=True)
    bugview = relationship(BugView,
                           backref=backref('bugviewtag_set',
                                           lazy='dynamic',
                                           cascade='all'))
    tag_id = Column(Integer, ForeignKey('tag.id'), primary_key=True)
    tag = relationship(Tag)

    def __init__(self, bugview, tag):
        self.bugview = bugview
        self.tag = tag

    def __repr__(self):
        return ('<BugViewTag: {0}@{1}:{2}>'
                .format(self.bugview.bug.id,
                        self.bugview.datechanged,
                        self.tag))


class BugTask(Base):
    __tablename__ = 'bugtask'
    __table_args__ = (UniqueConstraint('bug_id', 'project_id'),)
    id = Column(Integer, primary_key=True)
    bug_id = Column(Integer, ForeignKey('bug.id'))
    bug = relationship(Bug,
                       backref=backref('bugtask_set',
                                       lazy='dynamic'))
    project_id = Column(Integer, ForeignKey('project.id'))
    project = relationship(Project,
                           backref=backref('bugtask_set',
                                           lazy='dynamic',
                                           order_by=bug_id,
                                           cascade='all'))
    date_created = Column(DateTime)
    date_closed = Column(DateTime)
    date_reopened = Column(DateTime)

    def __init__(self, bug, project, date_created):
        self.bug = bug
        self.project = project
        self.date_created = date_created

    def __repr__(self):
        return ('<BugTask: {0}:{1}>'
                .format(self.bug.id, self.project.name))


class BugTaskView(Base):
    __tablename__ = 'bugtaskview'
    __table_args__ = (UniqueConstraint('datechanged', 'bugtask_id'),)
    id = Column(Integer, primary_key=True)
    datechanged = Column(DateTime, nullable=False)
    bugtask_id = Column(Integer,
                        ForeignKey('bugtask.id'),
                        nullable=False)
    bugtask = relationship(BugTask,
                           backref=backref('bugtaskview_set',
                                           lazy='dynamic',
                                           order_by=datechanged.desc(),
                                           cascade='all'))
    status_id = Column(Integer,
                       ForeignKey('status.id'),
                       nullable=False)
    status = relationship(Status)
    importance_id = Column(Integer,
                           ForeignKey('importance.id'),
                           nullable=False)
    importance = relationship(Importance)
    milestone_id = Column(Integer, ForeignKey('milestone.id'))
    milestone = relationship(Milestone,
                             backref=backref('bugtaskview_set',
                                             lazy='dynamic'))
    assignee_id = Column(Integer, ForeignKey('person.id'))
    assignee = relationship(Person,
                            backref=backref('bugtaskview_set',
                                            lazy='dynamic'))
    series_id = Column(Integer, ForeignKey('series.id'))
    series = relationship(Series)

    def __init__(self, datechanged, bugtask,
                 status, importance,
                 milestone=None, assignee=None,
                 project=None, series=None):
        self.datechanged = datechanged
        self.bugtask = bugtask
        self.status = status
        self.importance = importance
        self.milestone = milestone
        self.assignee = assignee
        self.project = project
        self.series = series

    def __repr__(self):
        return ('<BugTaskView: {0}:{1}@{2}>'
                .format(self.bugtask.bug.id,
                        self.bugtask.project.name,
                        self.datechanged))
