#!/usr/bin/env python
import re
import logging
from datetime import datetime
from urlparse import urlparse
from itertools import takewhile

from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound
from lazr.restfulclient.errors import NotFound

from . import ALL_BUG_STATUSES
from .model import Status, Importance, \
        Person, Series, Milestone, \
        Bug, BugTask, BugTaskView, \
        Tag, BugView, BugViewTag
from ..util import critical_section


class ProjectUpdater(object):
    """
    Update bug information for project
    """
    def __init__(self, launchpad, project):
        self.launchpad = launchpad
        self.session = Session.object_session(project)
        self.project = project

        logging.info('Getting project {0!r} from launchpad...'
                     .format(self.project.name))
        self.lp_project = launchpad.projects[self.project.name]

    def update(self, cancel_event=None, progressbar=None):
        """
        Update bug information for this project
        """
        logging.info('Searching tasks....')
        search_kwargs = {'status': ALL_BUG_STATUSES, }
        if self.project.date_updated:
            search_kwargs['modified_since'] = self.project.date_updated
        lp_tasks = self.lp_project.searchTasks(**search_kwargs)

        # Update all tasks one by one
        task_count = len(lp_tasks)
        task_count_length = len(str(task_count))

        # Update project timestamp
        # so that it can be used as a reference next time
        project = self.project
        project.date_updated = datetime.utcnow()

        for index, lp_task in enumerate(lp_tasks):
            if cancel_event and cancel_event.is_set():
                break

            lp_bug = lp_task.bug
            logging.info('{0:{1}}/{2} Getting bug task (#{3})...'
                         .format(index + 1, task_count_length,
                                 task_count, lp_bug.id))

            bug_updater = BugUpdater(self.project, self.launchpad,
                                     lp_bug, lp_task)
            bug_updater.update()

            if progressbar and not (cancel_event and cancel_event.is_set()):
                progress = float(index + 1) / task_count
                with critical_section():
                    progressbar.set_fraction(progress)
        else:
            # Make sure that even if no bug tasks are found
            # the progress bar is full
            if progressbar:
                with critical_section():
                    progressbar.set_fraction(1.0)

        updated = task_count > 0
        return updated


class BugUpdater(object):
    """
    Update bug information based on the activity records
    """
    ASSIGNEE_REGEX = re.compile(r'(?P<display_name>.*?) \((?P<name>.*?)\)')
    CHANGE_REGEX = re.compile(r'(?P<project>[^/:]+)(?:/(?P<series>[^:]+))?'
                              ': (?P<what>\w+)')

    OPENED_BUG_STATUSES = set([
            'New',
            'Incomplete',
            'Confirmed',
            'Triaged',
            'In Progress',
            ])
    CLOSED_BUG_STATUSES = set([
            'Invalid',
            'Opinion',
            "Won't Fix",
            'Fix Committed',
            'Fix Released',
            'Expired',
            ])

    def __init__(self, project, launchpad, lp_bug, lp_task):
        self.session = Session.object_session(project)
        self.project = project
        self.launchpad = launchpad
        self.lp_bug = lp_bug
        self.lp_task = lp_task

    def update(self):
        """
        Traverse bug activity records and create views for tasks
        """
        session = self.session
        project = self.project

        # Get bug from database
        lp_bug = self.lp_bug
        bug = session.query(Bug).get(lp_bug.id)
        if bug is None:
            bug = Bug(lp_bug.id, lp_bug.title)
            session.add(bug)

        # Get bug task current state
        taskview_kwargs = self._get_current_state(bug, project)
        bugtask = taskview_kwargs['bugtask']

        oldest_bugtaskview = bugtask.bugtaskview_set.first()
        if oldest_bugtaskview:
            # Process only new activity entries if bug task views
            # are found in the database
            oldest_datechanged = oldest_bugtaskview.datechanged

            def unkwown_entry(lp_entry):
                datechanged = lp_entry.datechanged.replace(tzinfo=None)
                return datechanged > oldest_datechanged
            lp_entries = takewhile(unkwown_entry,
                                   reversed(self.lp_bug.activity))
        else:
            # Process all activity entries if no bug task views
            # are found in the database
            lp_entries = reversed(self.lp_bug.activity)

        # Generate the new bug task views
        pending_lp_tags = lp_bug.tags
        for lp_entry in lp_entries:
            datechanged = lp_entry.datechanged.replace(tzinfo=None)
            whatchanged = lp_entry.whatchanged
            oldvalue = lp_entry.oldvalue
            newvalue = lp_entry.newvalue

            taskview_kwargs['datechanged'] = datechanged
            logging.debug('Processing entry: {0}, {1!r}, {2!r}, {3!r}'
                          .format(datechanged, whatchanged,
                                  newvalue, oldvalue))

            # Handle general attribute changes
            # by updating value in view
            match = self.CHANGE_REGEX.match(whatchanged)
            if match:
                self._handle_attribute_change(bugtask, taskview_kwargs,
                                              lp_entry, match)
                continue

            # Handle affected project changes
            # whatchanged == 'affects'
            # Not tracked anymore

            # Handle bug tag changes
            if whatchanged == 'tags':
                lp_tags = newvalue.split()
                pending_lp_tags = oldvalue.split()
                self._add_bugview(bug, datechanged, lp_tags)
                continue

            # Stop on bug task creation
            if whatchanged == 'bug task added':
                if project.name != newvalue:
                    logging.debug('Irrelevant entry '
                                  "(project name doesn't match):\n"
                                  '{0!r} != {1!r}'
                                  .format(newvalue, project.name))
                    continue

                self._add_bugtaskview(taskview_kwargs)
                logging.debug('Stopped at entry: {0}, {1!r}, {2!r}, {3!r} '
                              '(stop reason: bug task added)'
                              .format(datechanged, whatchanged,
                                      newvalue, oldvalue))
                break

            # Stop on bug creation
            if whatchanged == 'bug':
                self._add_bugtaskview(taskview_kwargs)
                logging.debug('Stopped at entry: {0}, {1!r}, {2!r}, {3!r} '
                              '(stop reason: bug created)'
                              .format(datechanged, whatchanged,
                                      newvalue, oldvalue))
                break

            # Warn about irrelevant entries (just in case)
            logging.debug('Irrelevant entry ({0!r} not being tracked):\n{1}'
                          .format(whatchanged, lp_entry))

        # Take care of tags that were assigned on bug creation
        if pending_lp_tags:
            self._add_bugview(bug, lp_bug.date_created, pending_lp_tags)

    def _get_current_state(self, bug, project):
        """
        Get current state for the bug task
        """
        session = self.session
        lp_task = self.lp_task

        # Get bug task
        try:
            task = (session.query(BugTask)
                    .filter_by(bug=bug, project=project).one())
        except NoResultFound:
            date_created = lp_task.date_created.replace(tzinfo=None)
            task = BugTask(bug, project, date_created)
            session.add(task)

        # Gather parameters for the initial bug task view
        status = (session.query(Status)
                  .filter_by(name=lp_task.status).one())

        importance = (session.query(Importance)
                      .filter_by(name=lp_task.importance).one())
        taskview_kwargs = {'datechanged': datetime.utcnow(),
                           'bugtask': task,
                           'status': status,
                           'importance': importance,
                           'milestone': None,
                           'assignee': None,
                           'series': None,
                           }

        milestone_str = getattr(lp_task.milestone, 'name', '')
        if milestone_str:
            try:
                milestone = (session.query(Milestone)
                             .filter_by(project=self.project,
                                        name=milestone_str)
                             .one())
            except NoResultFound:
                milestone = Milestone(milestone_str, project)
            taskview_kwargs['milestone'] = milestone

        lp_assignee = lp_task.assignee
        if lp_assignee:
            try:
                assignee = (session.query(Person)
                            .filter_by(name=lp_assignee.name).one())
            except NoResultFound:
                assignee = Person(lp_assignee.name,
                                  lp_assignee.display_name,
                                  lp_assignee.is_team)
            taskview_kwargs['assignee'] = assignee

        lp_target = lp_task.target
        lp_target_type = urlparse(lp_target.resource_type_link).fragment
        if lp_target_type == 'project':
            assert project.name == lp_target.name
        elif lp_target_type == 'project_series':
            assert project.name == lp_target.project.name
            try:
                series = (session.query(Series)
                          .filter_by(name=lp_target.name, project=project)
                          .one())
            except NoResultFound:
                series = Series(lp_target.name, project)
            taskview_kwargs['series'] = series
        else:
            raise ValueError('Unexpected target type: {0}'
                             .format(lp_target_type))
        return taskview_kwargs

    def _handle_attribute_change(self, bugtask, taskview_kwargs,
                                 lp_entry, match):
        """
        Take care of attribute change in bug activity
        """
        project = self.project
        session = self.session
        oldvalue = lp_entry.oldvalue
        newvalue = lp_entry.newvalue

        project_name, series_name, what = match.groups()
        logging.debug('Attribute changed: {0!r}, {1!r}, {2!r}'
                      .format(project_name, series_name, what))

        # Log any irrelevant entry properly
        if (project.name != project_name):
            logging.debug('Irrelevant entry '
                          "(project name doesn't match):\n"
                          '{0!r} != {1!r}'
                          .format(project_name, project.name))
        elif (taskview_kwargs['series']
              and taskview_kwargs['series'].name != series_name):
            logging.debug('Irrelevant entry '
                          "(series name doesn't match):"
                          '\n{0!r} != {1!r}'
                          .format(series_name,
                                  taskview_kwargs['series'].name))
        elif not what in taskview_kwargs:
            logging.debug('Irrelevant entry '
                          '({0!r} not being tracked):\n{1}'
                          .format(what, lp_entry))
        else:
            self._add_bugtaskview(taskview_kwargs)

            # Map strings to database records
            if what == 'status':
                if (newvalue in self.CLOSED_BUG_STATUSES
                    and oldvalue in self.OPENED_BUG_STATUSES):
                    date_closed = taskview_kwargs['datechanged']

                    # Update date_closed only for new entries
                    if (bugtask.date_closed is None
                        or bugtask.date_closed < date_closed):
                        logging.debug('Setting date_closed to: {0}'
                                      .format(date_closed))
                        bugtask.date_closed = date_closed

                        # Reset date_reopened
                        # when bugtask has been closed later
                        if (bugtask.date_reopened
                            and bugtask.date_reopened < date_closed):
                            logging.debug('Setting date_reopened to: {0}'
                                          .format(None))
                            bugtask.date_reopened = None

                elif (newvalue in self.OPENED_BUG_STATUSES
                      and oldvalue in self.CLOSED_BUG_STATUSES):
                    date_reopened = taskview_kwargs['datechanged']

                    # Update date_reopened only for new entries
                    if (bugtask.date_closed is None
                        or bugtask.date_closed < date_reopened):
                        logging.debug('Setting date_reopened to: {0}'
                                      .format(date_reopened))
                        bugtask.date_reopened = date_reopened

                old_status = (session.query(Status)
                              .filter_by(name=oldvalue).one())
                taskview_kwargs['status'] = old_status
            elif what == 'importance':
                old_importance = (session.query(Importance)
                                  .filter_by(name=oldvalue).one())
                taskview_kwargs['importance'] = old_importance
            elif what == 'assignee':
                if oldvalue is None:
                    taskview_kwargs['assignee'] = None
                else:
                    match = self.ASSIGNEE_REGEX.match(oldvalue)
                    if match:
                        display_name, name = match.groups()
                    else:
                        name = oldvalue
                        display_name = None

                    try:
                        old_assignee = (session.query(Person)
                                        .filter_by(name=name).one())
                    except NoResultFound:
                        try:
                            lp_assignee = self.launchpad.people(name)
                            old_assignee = Person(lp_assignee.name,
                                                  lp_assignee.display_name,
                                                  lp_assignee.is_team)
                        except NotFound:
                            logging.warning('Assignee not found '
                                            'in launchpad: {0}'
                                            .format(name))
                            old_assignee = Person(name, display_name)

                    taskview_kwargs['assignee'] = old_assignee
            elif what == 'milestone':
                if oldvalue is None:
                    taskview_kwargs['milestone'] = None
                else:
                    try:
                        old_milestone = \
                                (session.query(Milestone)
                                 .filter_by(name=oldvalue, project=project)
                                 .one())
                    except NoResultFound:
                        old_milestone = Milestone(oldvalue, project)
                    taskview_kwargs['milestone'] = old_milestone
            else:
                raise ValueError('Unexpected value type: {0!r}'
                                 .format(what))
            logging.debug('{0!r} updated from {1!r} to {2!r}'
                          .format(what, newvalue, oldvalue))

    def _add_bugview(self, bug, datechanged, lp_tags):
        """
        Create tag, bugview and bugviewtag records and add them to database
        """
        session = self.session
        try:
            # Check if bugview exists to avoid creating duplicated entries
            bugview = (session.query(BugView)
                       .filter_by(datechanged=datechanged, bug=bug).one())
        except NoResultFound:
            session = self.session
            tags = []
            for lp_tag in lp_tags:
                try:
                    tag = session.query(Tag).filter_by(name=lp_tag).one()
                except NoResultFound:
                    tag = Tag(lp_tag)
                    session.add(tag)
                tags.append(tag)

            bugview = BugView(datechanged, bug)
            session.add(bugview)
            for tag in tags:
                session.add(BugViewTag(bugview, tag))

    def _add_bugtaskview(self, bugtaskview_kwargs):
        """
        Create bug task view record and add it to database
        """
        session = self.session
        datechanged = bugtaskview_kwargs['datechanged']
        bugtask = bugtaskview_kwargs['bugtask']
        try:
            # When the timestap matches one that already exists,
            # don't create any bug task view since all changes
            # must be joined together in the next timestamp
            bugtaskview = (session.query(BugTaskView)
                           .filter_by(datechanged=datechanged,
                                      bugtask=bugtask).one())
        except NoResultFound:
            bugtaskview = BugTaskView(**bugtaskview_kwargs)
            session.add(bugtaskview)
