"""
Application dialogs
"""
import logging
import threading
import time

from gi.repository import Gtk

from .uifile import UIFile
from .db.model import Project
from .util import critical_section


class ProjectDialog(UIFile):
    """
    Dialog used to get a new launchpad project name
    """
    def __init__(self, parent, launchpad, session):
        super(ProjectDialog, self).__init__('dialog/add_project.ui')
        self.launchpad = launchpad
        self.session = session
        self.dialog.set_transient_for(parent)
        self.dialog.show_all()

    def run(self):
        """
        Run dialog and return project name
        """
        response_id = self.dialog.run()
        if response_id == Gtk.ResponseType.OK:
            project = self.project_entry.get_text()
        else:
            project = None
        self.dialog.destroy()
        return project

    def project_changed_cb(self, entry):
        """
        Check that project name isn't already in the database
        and exists in launchpad
        """
        project_name = entry.get_text()

        project_is_valid = bool(project_name)
        if project_is_valid:
            project = (self.session.query(Project)
                       .filter_by(name=project_name).first())
            project_is_valid = project is None

        if project_is_valid:
            try:
                self.launchpad.projects[project_name]
            except KeyError:
                project_is_valid = False

        if project_is_valid:
            self.ok_button.set_sensitive(True)
            self.dialog.set_default_response(Gtk.ResponseType.OK)
        else:
            self.ok_button.set_sensitive(False)
            self.dialog.set_default_response(Gtk.ResponseType.CANCEL)


class DbUpdateDialog(UIFile):
    """
    Dialog used to display update progress
    """
    def __init__(self, parent):
        super(DbUpdateDialog, self).__init__('dialog/update.ui')
        self.dialog.set_transient_for(parent)
        self.dialog.show_all()
        self.db_updated = False

    def run(self, project_updaters):
        """
        Update project information
        """
        cancel_event = threading.Event()
        thread = threading.Thread(target=self.update_target,
                                  args=(project_updaters, cancel_event))
        thread.start()
        response_id = self.dialog.run()
        if response_id != Gtk.ResponseType.OK:
            cancel_event.set()
            thread.join()
        self.dialog.destroy()
        return response_id, self.db_updated

    def update_target(self, project_updaters, cancel_event):
        """
        Update bug information in a separate thread
        """
        project_count = len(project_updaters)
        project_count_length = len(str(project_count))

        def run_update(index, project_updater):
            message = ('{0:{1}}/{2} Updating bug tasks for {3!r}...'
                       .format(index + 1, project_count_length,
                               project_count, project_updater.project.name))
            logging.info(message)

            project_updated = project_updater.update(cancel_event,
                                                     self.bugs_progressbar)

            if not cancel_event.is_set():
                progress = float(index + 1) / project_count
                with critical_section():
                    self.projects_progressbar.set_fraction(progress)

            return project_updated

        self.db_updated = any([run_update(index, project_updater)
                               for index, project_updater
                               in enumerate(project_updaters)
                               if not cancel_event.is_set()])

        if cancel_event.is_set():
            logging.info('Update cancelled by user')
        else:
            self.ok_button.set_sensitive(True)
            logging.info('Update completed')


class ProgressDialog(UIFile):
    """
    Dialog used to give feedback to the user
    about the progress of some action
    """
    def __init__(self, parent, title):
        super(ProgressDialog, self).__init__('dialog/progress.ui')
        self.dialog.set_transient_for(parent)
        self.dialog.set_title(title)
        self.dialog.show_all()

    def delete_event_cb(self, dialog, event):
        """
        Ignore any attempt to close the dialog window
        """
        return True

    def run(self, worker_function):
        """
        Populate tree stores
        """
        def target():
            worker_function(self.progressbar)
            # Small delay just to let the user see
            # the progress bar complete before destroying the dialog
            time.sleep(0.5)
            self.dialog.destroy()

        thread = threading.Thread(target=target)
        thread.start()
