"""
``Gtk.Notebook`` specialized class
"""
from gi.repository import Gtk

from .tab import ProjectTab, MilestoneTab, PersonTab, BugTab
from ..db.model import Project, Milestone, Person, Bug


class Notebook(Gtk.Notebook):
    """
    ``Gtk.Notebook`` wrapper that takes care of opening/closing tabs
    that display the contents of a row in any of the application
    models.

    New tabs are opened when a row is activated in the treeview and
    closed when the tab close button is clicked or when the row that
    it displays is removed from the model
    """
    def __init__(self, application):
        """
        Initialize notebook object
        """
        Gtk.Notebook.__init__(self)
        self.set_scrollable(True)
        self.tabs = []
        self.application = application

        for object_type in ('projects', 'people', 'bugs'):
            treeview_name = '{0}_treeview'.format(object_type)
            treeview = getattr(application, treeview_name)
            treeview.connect('row-activated', self.row_activated_cb)

    def find_tab(self, tab_id):
        """
        Look for an already opened tab displaying a row
        """
        for tab in self.tabs:
            if tab.id == tab_id:
                return tab

        return None

    def open_tab(self, tab_id, tab_cls, tab_args):
        """
        Display information in a tab.

        A new tab will be created if the id provided
        doesn't match any existing tab. Otherwise,
        the tab will be just focused.
        """
        # If row is displayed by an already opened tab,
        # then just focus it
        tab = self.find_tab(tab_id)
        if tab is not None:
            page_num = self.page_num(tab.page)
            assert page_num != -1, 'Tab not found'
            self.set_current_page(page_num)
            return tab

        # If tab_id doesn't match any tab, then create a new tab
        tab = tab_cls(*tab_args)

        # Append the new tab to notebook
        page_number = self.append_page(tab.page, tab.header)
        self.set_tab_reorderable(tab.page, True)
        self.set_current_page(page_number)
        self.tabs.append(tab)
        return tab

    def close_tab(self, tab):
        """
        Close an existing tab
        """
        page_num = self.page_num(tab.page)
        assert page_num != -1, 'Tab not found'
        self.remove_page(page_num)

        self.tabs.remove(tab)
        tab.destroy()

    def find_and_close_tab(self, tab_id):
        """
        Try to find a tab and close it if it exists
        """
        tab = self.find_tab(tab_id)
        if tab:
            self.close_tab(tab)

    def close_all_tabs(self):
        """
        Close all tabs
        """
        tabs = self.tabs[:]
        for tab in tabs:
            self.close_tab(tab)

    def refresh_all_tabs(self):
        """
        Refresh all tabs
        """
        tabs = self.tabs[:]
        for tab in tabs:
            tab.refresh()

    def row_activated_cb(self, treeview, path, view_column):
        """
        Open tab to display database record
        for the activated row in the left treeview
        """
        store = treeview.get_model()
        store_name = store.get_name()
        store_row = store[path]

        if store_name == 'projects':
            depth = path.get_depth()
            if depth == 1:
                db_model = Project
                tab_cls = ProjectTab
            elif depth == 2:
                db_model = Milestone
                tab_cls = MilestoneTab
            else:
                raise ValueError('Unexpected depth in projects row: {0}'
                                 .format(depth))
        elif store_name == 'people':
            db_model = Person
            tab_cls = PersonTab
        elif store_name == 'bugs':
            db_model = Bug
            tab_cls = BugTab
        else:
            raise ValueError('Unexpected store name: {0}'.format(store_name))

        session = self.application.db.session
        db_row = session.query(db_model).get(store_row.id)
        tab_id = tab_cls.get_id(db_row)
        self.open_tab(tab_id, tab_cls, (db_row, self.application))
