"""
Bug tab
"""
from .tab import DBTab
from ...db.model import Project, Bug, BugTask, Tag, BugViewTag
from ...tree.model import TreeStore
from ...tree.model import TreeStoreColumnMetadata as ColumnMetadata


class BugTab(DBTab):
    """
    `Tab` that displays launchpad bug information
    """
    ICON_BASENAME = 'bug.png'

    def __init__(self, bug, application):
        super(BugTab, self).__init__(bug, application)
        self.bug = bug
        self.title = '{0}'.format(bug.id)
        self.bug_id_label.set_text('Bug #{0}'.format(bug.id))
        self.bug_title_textbuffer.set_text(bug.title.strip())

        store = TreeStore('bugtags',
                          (ColumnMetadata('datechanged', str),
                           ColumnMetadata('tags', str)))
        self.bugtags_store = store
        self.bugtags_treeview.set_model(store)

        store = TreeStore('bugtasks',
                          (ColumnMetadata('project', str),
                           ColumnMetadata('datechanged', str),
                           ColumnMetadata('status', str),
                           ColumnMetadata('importance', str),
                           ColumnMetadata('assignee', str),
                           ColumnMetadata('milestone', str),
                          ))
        self.bugtasks_store = store
        self.bugtasks_treeview.set_model(store)

        self.refresh()

    def refresh(self):
        """
        Refresh tab content
        """
        # Make sure that bug still exists in database
        # Otherwise, close tab
        session = self.application.db.session
        bug = session.query(Bug).get(self.bug.id)
        if bug is None:
            self.close()
            return

        # Bug tags
        bugtags_store = self.bugtags_store
        bugtags_store.clear()
        bugview_set = bug.bugview_set
        bugview_count = bugview_set.count()
        bugtags_label = self.bugtags_label
        bugtags_label.set_text('{0} bug tag change{1} found'
                               .format(bugview_count,
                                       '' if bugview_count == 1 else 's'))
        for bugview in bugview_set:
            datechanged = bugview.datechanged.strftime('%c')
            tags = (session.query(Tag)
                    .join(BugViewTag).filter_by(bugview=bugview)
                    .order_by(Tag.name).all())
            bugviewtags = ', '.join([tag.name for tag in tags])
            bugtags_store.append(None, (datechanged, bugviewtags))

        # Bug tasks
        bugtasks_store = self.bugtasks_store
        bugtasks_store.clear()
        bugtask_set = (session.query(BugTask)
                       .join(Project.bugtask_set)
                       .filter_by(bug=bug)
                       .order_by(Project.name))
        bugtask_count = bugtask_set.count()
        bugtasks_label = self.bugtasks_label
        bugtasks_label.set_text('{0} bug task{1} found'
                                .format(bugtask_count,
                                        '' if bugtask_count == 1 else 's'))
        for bugtask in bugtask_set:
            project = bugtask.project.name
            row_tuple = (project, '', '', '', '', '')
            bugtask_iter = bugtasks_store.append(None, row_tuple)
            bugtaskview_set = bugtask.bugtaskview_set

            for bugtaskview in bugtaskview_set:
                datechanged = bugtaskview.datechanged.strftime('%c')
                status = bugtaskview.status.name
                importance = bugtaskview.importance.name
                if bugtaskview.assignee:
                    assignee = bugtaskview.assignee.name
                else:
                    assignee = ''
                if bugtaskview.milestone:
                    milestone = bugtaskview.milestone.name
                else:
                    milestone = ''
                row_tuple = ('', datechanged, status, importance,
                             assignee, milestone)
                bugtasks_store.append(bugtask_iter, row_tuple)

        # Display complete bug information by default
        self.bugtasks_treeview.expand_all()
