"""
Home tab
"""
from datetime import datetime, timedelta

from gi.repository import Gtk, Gdk
from sqlalchemy.orm.exc import NoResultFound

from .tab import Tab
from .person import PersonTab
from .unverified import UnverifiedTab
from .age import AgeTab
from .regression import RegressionTab
from ...db.model import Project, Person, Bug, BugTask, Status


class HomeTab(Tab):
    """
    `Tab` that displays summary information
    """
    id = 'home'

    def __init__(self, application):
        """
        Get launchpad user information
        and display other summary information
        """
        icon_image = Gtk.Image()
        icon_image.set_from_stock(Gtk.STOCK_HOME, Gtk.IconSize.SMALL_TOOLBAR)
        super(HomeTab, self).__init__('home.ui', icon_image, application)
        self.title = 'Home'

        launchpad = application.launchpad
        lp_person = launchpad.me
        name = lp_person.name
        display_name = lp_person.display_name
        self.launchpad_user_label.set_text(name)

        session = application.db.session
        # Make user that user is in the database
        try:
            person = session.query(Person).filter_by(name=name).one()
        except NoResultFound:
            is_team = lp_person.is_team
            person = Person(name, display_name, is_team)
            session.add(person)

        self.person = person

        self.refresh()

    def refresh(self):
        """
        Refresh tab content
        """
        session = self.application.db.session

        # Get number of projects
        project_count = session.query(Project).count()
        self.project_count_label.set_text(str(project_count))

        # Get number of people
        person_count = session.query(Person).count()
        self.person_count_label.set_text(str(person_count))

        # Get number of bug tasks
        bug_count = session.query(Bug).count()
        self.bug_count_label.set_text(str(bug_count))

        bugs = session.query(Bug).all()

        # Get number of unverified bug tasks
        fix_committed = (session.query(Status)
                         .filter_by(name='Fix Committed').one())
        fix_released = (session.query(Status)
                        .filter_by(name='Fix Released').one())
        fixed = (fix_committed, fix_released)
        unverified_bugs = []
        for bug in bugs:
            bug_fixed = any(bugtask.bugtaskview_set.first().status in fixed
                            for bugtask in bug.bugtask_set.all())
            if bug_fixed:
                bugview = bug.bugview_set.first()
                if bugview:
                    tags = [bugviewtag.tag.name
                            for bugviewtag in bugview.bugviewtag_set.all()]
                    bug_verified = 'cqa-verified' in tags
                    if not bug_verified:
                        unverified_bugs.append(bug)
                else:
                    unverified_bugs.append(bug)
        self.unverified_count_label.set_text('{0}'
                                             .format(len(unverified_bugs)))
        self.unverified_bugs = unverified_bugs

        # Calculate average bug age
        bugtask_ages = []
        for bug in bugs:
            bugtasks = bug.bugtask_set.all()
            for bugtask in bugtasks:
                date_created = bugtask.date_created
                date_closed = bugtask.date_closed
                date_reopened = bugtask.date_reopened
                if date_closed and not date_reopened:
                    age = date_closed - date_created
                else:
                    age = datetime.now() - date_created
                bugtask_ages.append(age)
        if len(bugtask_ages) > 0:
            bugtask_age_average = (sum(bugtask_ages, timedelta())
                                   / len(bugtask_ages))
            self.age_average_label.set_text(str(bugtask_age_average.days))
        else:
            self.age_average_label.set_text('0')

        # Calculate regressions
        regression_count = (session.query(BugTask)
                            .filter(BugTask.date_reopened != None)
                            .join(Bug).order_by(Bug.id)
                            .join(Project).order_by(Project.name)
                            .count())
        self.regressions_count_label.set_text('{0}'.format(regression_count))

    def projects_cb(self, widget, event):
        """
        Select projects tab
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            notebook = self.application.left_notebook
            notebook.set_current_page(0)

    def people_cb(self, widget, event):
        """
        Select people tab
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            notebook = self.application.left_notebook
            notebook.set_current_page(1)

    def bugs_cb(self, widget, event):
        """
        Select bugs tab
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            notebook = self.application.left_notebook
            notebook.set_current_page(2)

    def unverified_cb(self, widget, event):
        """
        Display unverified bugs
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            notebook = self.parent_notebook
            notebook.open_tab(UnverifiedTab.id,
                              UnverifiedTab, (self.application,))

    def age_cb(self, widget, event):
        """
        Display bug tasks age
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            notebook = self.parent_notebook
            notebook.open_tab(AgeTab.id,
                              AgeTab, (self.application,))

    def regressions_cb(self, widget, event):
        """
        Display regressions found
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            notebook = self.parent_notebook
            notebook.open_tab(RegressionTab.id,
                              RegressionTab, (self.application,))

    def lp_user_activated_cb(self, widget, event):
        """
        Open person tab when launchpad user name is activated
        """
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            tab_id = PersonTab.get_id(self.person)
            notebook = self.parent_notebook
            notebook.open_tab(tab_id,
                              PersonTab, (self.person, self.application))
