"""
Project tab
"""
from .tab import DBTab
from ...tree.model import TreeStore
from ...tree.model import TreeStoreColumnMetadata as ColumnMetadata
from ...db.model import Milestone, Tag, BugViewTag


class MilestoneTab(DBTab):
    """
    `Tab` that displays launchpad milestone information
    """
    ICON_BASENAME = 'milestone.png'

    def __init__(self, milestone, application):
        super(MilestoneTab, self).__init__(milestone, application)
        self.milestone = milestone
        project = milestone.project

        title = ('{0} - {1}'.format(project.name, milestone.name))
        self.title = title
        self.milestone_name_label.set_text(title)

        store = TreeStore('bugtasks',
                          (ColumnMetadata('bug_id', int),
                           ColumnMetadata('status', str),
                           ColumnMetadata('importance', str),
                           ColumnMetadata('assignee', str),
                           ColumnMetadata('tags', str),
                          ))
        self.bugtasks_store = store
        self.bugtasks_treeview.set_model(store)
        self.refresh()

    def refresh(self):
        """
        Refresh tab content
        """
        # Make sure that milestone still exists in database
        # Otherwise, close tab
        session = self.application.db.session
        milestone = session.query(Milestone).get(self.milestone.id)
        if milestone is None:
            self.close()
            return
        store = self.bugtasks_store
        store.clear()
        milestone = self.milestone

        bugtasks = set(bugtaskview.bugtask
                       for bugtaskview in milestone.bugtaskview_set)
        bugtasks = [bugtask for bugtask in bugtasks
                    if bugtask.bugtaskview_set.first().milestone == milestone]
        bugtasks = sorted(bugtasks, key=lambda bugtask: bugtask.bug_id)
        self.bugtasks_label.set_text('{0} bug tasks found'
                                      .format(len(bugtasks)))

        for bugtask in bugtasks:
            bug = bugtask.bug
            bugview = bug.bugview_set.first()
            if bugview is None:
                tags = ''
            else:
                bugviewtag_set = (session.query(Tag).join(BugViewTag)
                                  .filter(BugViewTag.bugview == bugview)
                                  .order_by(Tag.name).all())
                tags = ', '.join([bugviewtag.name
                                  for bugviewtag in bugviewtag_set])

            bug_id = bugtask.bug_id
            bugtaskview = bugtask.bugtaskview_set.first()
            status = bugtaskview.status.name
            importance = bugtaskview.importance.name
            if bugtaskview.assignee:
                assignee = bugtaskview.assignee.name
            else:
                assignee = ''
            store.append(None, (bug_id, status, importance, assignee, tags))
