"""
Person tab
"""
from .tab import DBTab
from ...tree.model import TreeStore
from ...tree.model import TreeStoreColumnMetadata as ColumnMetadata
from ...db.model import Person, Tag, BugViewTag
from ...uifile import Image


class PersonTab(DBTab):
    """
    `Tab` that displays launchpad person information
    """
    def __init__(self, person, application):
        super(PersonTab, self).__init__(person, application)
        self.person = person
        self.title = person.name
        self.person_name_label.set_text('{0} ({1})'
                                        .format(person.name,
                                                person.display_name))

        store = TreeStore('bugtasks',
                          (ColumnMetadata('bug_id', int),
                           ColumnMetadata('status', str),
                           ColumnMetadata('importance', str),
                           ColumnMetadata('project', str),
                           ColumnMetadata('milestone', str),
                           ColumnMetadata('tags', str),
                          ))
        self.bugtasks_store = store
        self.bugtasks_treeview.set_model(store)
        self.refresh()

    def refresh(self):
        """
        Refresh tab content
        """
        # Make sure that person still exists in database
        # Otherwise, close tab
        session = self.application.db.session
        person = session.query(Person).get(self.person.id)
        if person is None:
            self.close()
            return

        store = self.bugtasks_store
        store.clear()

        bugtasks = set(bugtaskview.bugtask
                       for bugtaskview in person.bugtaskview_set)
        bugtasks = [bugtask for bugtask in bugtasks
                    if bugtask.bugtaskview_set.first().assignee == person]
        bugtasks = sorted(bugtasks, key=lambda bugtask: bugtask.bug_id)
        self.bugtaskviews_label.set_text('{0} assigned bug tasks found'
                                         .format(len(bugtasks)))

        for bugtask in bugtasks:
            bug = bugtask.bug
            bugview = bug.bugview_set.first()
            if bugview is None:
                tags = ''
            else:
                bugviewtag_set = (session.query(Tag).join(BugViewTag)
                                  .filter(BugViewTag.bugview == bugview)
                                  .order_by(Tag.name).all())
                tags = ', '.join([bugviewtag.name
                                  for bugviewtag in bugviewtag_set])

            bug_id = bugtask.bug_id
            bugtaskview = bugtask.bugtaskview_set.first()
            status = bugtaskview.status.name
            importance = bugtaskview.importance.name
            project = bugtask.project.name
            if bugtaskview.milestone:
                milestone = bugtaskview.milestone.name
            else:
                milestone = ''
            store.append(None,
                        (bug_id, status, importance, project, milestone, tags))

    def _get_icon_image(self, person):
        """
        Return the filename for the icon to be displayed in the tab
        """
        if person.is_team:
            basename = 'team.png'
        else:
            basename = 'person.png'
        image = Image(basename)
        return image
