"""
Plot tab
"""
from operator import attrgetter
from collections import defaultdict, Counter
from datetime import datetime

from dateutil.relativedelta import relativedelta

import matplotlib as mpl
from matplotlib.figure import Figure
from matplotlib.backends.backend_gtk3agg import \
    FigureCanvasGTK3Agg as FigureCanvas
from matplotlib.backends.backend_gtk3 import \
    NavigationToolbar2GTK3 as NavigationToolbar

from .tab import Tab
from ...uifile import Image
from ...db.model import Project, BugTaskView
from ...dialog import ProgressDialog
from ...util import critical_section


class PlotTab(Tab):
    """
    `Tab` used to plot bug information statistics for a project
    """
    ICON_BASENAME = 'plot.png'
    TIME_FORMAT = '%Y-%m-%d %H:%M:%S'

    def __init__(self, project, application):
        icon_image = Image(self.ICON_BASENAME)
        super(PlotTab, self).__init__('plot.ui', icon_image, application)
        self.title = project.name
        self.project_label.set_text(project.name)

        figure = Figure()
        canvas = FigureCanvas(figure)
        self.plot_vbox.pack_start(canvas, expand=True, fill=True, padding=0)

        toolbar = NavigationToolbar(canvas, application.window)
        self.plot_vbox.pack_start(toolbar, expand=False, fill=True, padding=0)
        self.plot_vbox.show_all()

        self.project = project
        self.id = self.get_id(project.name)
        self.figure = figure

        self.refresh()

    @staticmethod
    def get_id(project_name):
        """
        Return tab id based on the project name used to display data
        """
        return 'plot-{0}'.format(project_name)

    def refresh(self):
        """
        Refresh tab content
        """
        # Make sure that project still exists in database
        # Otherwise, close tab
        session = self.application.db.session
        project = session.query(Project).get(self.project.id)
        if project is None:
            self.close()
            return

        # Update people and milestones involved in this project
        bugtaskviews = [bugtask.bugtaskview_set.first()
                        for bugtask in project.bugtask_set.all()]
        for attribute_name in ['assignee', 'milestone']:
            db_rows = set()
            for bugtaskview in bugtaskviews:
                db_row = getattr(bugtaskview, attribute_name)
                if db_row:
                    db_rows.add(db_row)
            db_rows = sorted(db_rows, key=attrgetter('name'))
            store = getattr(self, '{0}_store'.format(attribute_name))
            store.clear()
            for db_row in db_rows:
                store.append((False, db_row.name))
            store.append((False, 'Unassigned'))

        # Keep track of changes in filters to know when to draw figure
        self.draw_figure = True
        if self.notebook.get_current_page() == 1:
            self._plot()

    def attribute_changed_cb(self, combobox):
        """
        Update filter notebook when a new attribute is selected
        """
        index = combobox.get_active()
        self.filter_notebook.set_current_page(index)
        self.draw_figure = True

    def sampling_changed_cb(self, widget):
        """
        Update dates according to selected sampling rate
        """
        self.draw_figure = True

    def _generate_dates(self):
        """
        Generate array with all sampling dates
        """
        date_created = self.project.date_created
        from_datetime = datetime(day=date_created.day,
                                 month=date_created.month,
                                 year=date_created.year)

        # Use tomorrow as final date to make sure
        # that latest changes are included in the plot
        tomorrow = datetime.now() + relativedelta(days=1)
        to_datetime = datetime(day=tomorrow.day,
                               month=tomorrow.month,
                               year=tomorrow.year)

        # Get delta interval based on selected sampling
        sampling_amount = self.sampling_amount_spinbutton.get_value_as_int()
        sampling_unit_index = self.sampling_unit_combobox.get_active()
        sampling_unit = (self.sampling_unit_store[sampling_unit_index][0]
                         .lower())
        delta_kwargs = {sampling_unit: sampling_amount}
        delta = relativedelta(**delta_kwargs)

        dates = []
        date = from_datetime
        while date <= to_datetime:
            dates.append(date)
            date += delta
        return dates

    def _filter_toggled(self, store, path):
        """
        Update filter
        """
        selected, _ = store[path]
        new_selected = not selected
        iterator = store.get_iter(path)
        store.set_value(iterator, 0, new_selected)

        self.draw_figure = True

    def status_filter_toggled_cb(self, renderer, path):
        """
        Update status filter
        """
        store = self.status_store
        self._filter_toggled(store, path)

    def importance_filter_toggled_cb(self, renderer, path):
        """
        Update importance filter
        """
        store = self.importance_store
        self._filter_toggled(store, path)

    def assignee_filter_toggled_cb(self, renderer, path):
        """
        Update assignee filter
        """
        store = self.assignee_store
        self._filter_toggled(store, path)

    def milestone_filter_toggled_cb(self, renderer, path):
        """
        Update milestone filter
        """
        store = self.milestone_store
        self._filter_toggled(store, path)

    def page_switched_cb(self, notebook, page, page_num):
        """
        Update plot window when tab is selected
        """
        if page_num != 1 or not self.draw_figure:
            return
        self._plot()

    def _plot(self):
        """
        Plot selected data in figure
        """
        session = self.application.db.session

        # Get which attribute should be plotted
        attribute_index = self.attribute_combobox.get_active()
        attribute = self.attribute_store[attribute_index][0].lower()

        # Update series store depending on attribute being plotted
        series_store = self.series_store
        series_store.clear()
        attribute_store_name = '{0}_store'.format(attribute)
        attribute_store = getattr(self, attribute_store_name)
        for _, name in attribute_store:
            series_store.append((True, name))

        # Get filters
        all_attributes = ['status', 'importance', 'assignee', 'milestone']
        attr_filters = [attr for attr in all_attributes
                        if attr != attribute]
        filters = defaultdict(list)
        for attr_filter in attr_filters:
            store_name = '{0}_store'.format(attr_filter)
            store = getattr(self, store_name)
            for row in store:
                selected, name = row
                if selected:
                    if name == 'Unassigned':
                        value = None
                    else:
                        value = name
                    filters[attr_filter].append(value)

        # Gather number of bugs that satisfy the given criteria
        def bugtaskview_matches(bugtaskview):
            def getattr_name(attr_filter):
                attr = getattr(bugtaskview, attr_filter)
                name = attr.name if attr else None
                return name

            return all(getattr_name(attr_filter) in names
                       for attr_filter, names in filters.items()
                       if names)
        bugtasks = self.project.bugtask_set.all()
        figure = self.figure
        figure.clear()
        dates = self._generate_dates()
        counts = []
        lines = {}
        series = {}

        def target(progressbar):
            """
            Gather data for all the dates to be plotted
            """
            dates_count = len(dates)
            for index, date in enumerate(dates):
                progress = float(index + 1) / dates_count
                with critical_section():
                    progressbar.set_fraction(progress)

                bugtaskviews = [(session.query(BugTaskView)
                                 .filter_by(bugtask=bugtask)
                                 .filter(BugTaskView.datechanged <= date)
                                 .order_by(BugTaskView.datechanged.desc())
                                 .first())
                                 for bugtask in bugtasks]
                bugtaskviews = filter(None, bugtaskviews)
                attributes = [getattr(bugtaskview, attribute)
                              for bugtaskview in bugtaskviews
                              if bugtaskview_matches(bugtaskview)]
                attribute_names = [attr.name for attr in attributes if attr]
                counts.append(Counter(attribute_names))

            with critical_section():
                progressbar.set_fraction(1.0)
                figure.add_subplot(1, 1, 1)
                axes = figure.get_axes()[0]

            for selected, name in series_store:
                if selected:
                    series_counts = [count[name] for count in counts]
                    with critical_section():
                        line, = axes.plot_date(dates, series_counts, '-')
                    series[name] = series_counts
                    lines[name] = line

            with critical_section():
                # y axis ticks are always integer values
                (axes.yaxis
                 .set_major_locator(mpl.ticker.MaxNLocator(integer=True)))

                if self.legend_checkbutton.get_active():
                    labels = [name for _, name in series_store]
                    axes.legend(labels, loc='best')

                title = self.title_entry.get_text()
                axes.set_title(title)
                label = self.x_axis_entry.get_text()
                axes.set_xlabel(label)
                label = self.y_axis_entry.get_text()
                axes.set_ylabel(label)

                # Display dates properly
                axes.fmt_xdate = mpl.dates.DateFormatter(self.TIME_FORMAT)
                figure.autofmt_xdate()

                self.figure.canvas.draw()

        dialog = ProgressDialog(self.application.window,
                                '{0} plot update'.format(self.project.name))
        dialog.run(target)
        self.dates = dates
        self.series = series
        self.lines = lines
        self.draw_figure = False

    def title_changed_cb(self, widget):
        """
        Update figure title
        """
        new_title = widget.get_text()
        axes = self.figure.get_axes()[0]
        axes.set_title(new_title)
        self.figure.canvas.draw()

    def x_axis_changed_cb(self, widget):
        """
        Update x axis label
        """
        new_label = widget.get_text()
        axes = self.figure.get_axes()[0]
        axes.set_xlabel(new_label)
        self.figure.canvas.draw()

    def y_axis_changed_cb(self, widget):
        """
        Update y axis label
        """
        new_label = widget.get_text()
        axes = self.figure.get_axes()[0]
        axes.set_ylabel(new_label)
        self.figure.canvas.draw()

    def legend_toggled_cb(self, button):
        """
        Display/Hide legend
        """
        active = button.get_active()
        if active:
            self._display_legend()
        else:
            self._hide_legend()
        self.figure.canvas.draw()

    def _display_legend(self):
        """
        Display legend
        """
        labels = [name for _, name in self.series_store
                  if name in self.lines]
        if labels:
            axes = self.figure.get_axes()[0]
            axes.legend(labels, loc='best')
        else:
            # Hide legend
            # if there are no lines to display
            self._hide_legend()

    def _hide_legend(self):
        """
        Hide leyend
        """
        axes = self.figure.get_axes()[0]
        axes.legend_ = None

    def series_toggled_cb(self, renderer, path):
        """
        Display/Hide selected series
        """
        store = self.series_store
        selected, name = store[path]
        new_selected = not selected
        iterator = store.get_iter(path)
        store.set_value(iterator, 0, new_selected)

        axes = self.figure.get_axes()[0]
        if new_selected:
            # Remove all lines and plot them again
            # to keep legend ordering
            axes.lines = []
            for selected, name in store:
                if selected:
                    line, = axes.plot_date(self.dates,
                                           self.series[name], '-')
                    self.lines[name] = line
        else:
            line = self.lines[name]
            axes.lines.remove(line)
            del self.lines[name]

        # Update legend
        if self.legend_checkbutton.get_active():
            self._hide_legend()
            self._display_legend()

        self.figure.canvas.draw()
        self._update_series_selection_buttons()

    def select_all_series_cb(self, widget):
        """
        Select all series available
        """
        self._set_all_series(True)

    def deselect_all_series_cb(self, widget):
        """
        Deselect all series available
        """
        self._set_all_series(False)

    def _set_all_series(self, value):
        """
        Set all series to a given value
        """
        store = self.series_store
        for row in store:
            row[0] = value

        axes = self.figure.get_axes()[0]
        axes.lines = []
        if value:
            for selected, name in store:
                line, = axes.plot_date(self.dates,
                                       self.series[name], '-')
                self.lines[name] = line
        else:
            self.lines = {}

        # Update legend
        if self.legend_checkbutton.get_active():
            self._hide_legend()
            self._display_legend()

        self.figure.canvas.draw()
        self._update_series_selection_buttons()

    def _update_series_selection_buttons(self):
        """
        Update series selection buttons sensitivity
        """
        store = self.series_store
        rows_selected = [selected for selected, _ in store]
        all_rows_selected = all(rows_selected)
        any_row_selected = any(rows_selected)
        self.select_all_button.set_sensitive(not all_rows_selected)
        self.deselect_all_button.set_sensitive(any_row_selected)
