"""
Project tab
"""
from .tab import DBTab
from .plot import PlotTab
from ...tree.model import TreeStore
from ...tree.model import TreeStoreColumnMetadata as ColumnMetadata
from ...db.model import Project, Tag, BugViewTag


class ProjectTab(DBTab):
    """
    `Tab` that displays launchpad project information
    """
    ICON_BASENAME = 'project.png'

    def __init__(self, project, application):
        super(ProjectTab, self).__init__(project, application)
        self.project = project
        self.title = project.name
        self.project_name_label.set_text(project.name)

        store = TreeStore('bugtasks',
                          (ColumnMetadata('bug_id', int),
                           ColumnMetadata('status', str),
                           ColumnMetadata('importance', str),
                           ColumnMetadata('assignee', str),
                           ColumnMetadata('milestone', str),
                           ColumnMetadata('tags', str),
                          ))
        self.bugtasks_store = store
        self.bugtasks_treeview.set_model(store)
        self.refresh()

    def refresh(self):
        """
        Refresh tab content
        """
        # Make sure that project still exists in database
        # Otherwise, close tab
        session = self.application.db.session
        project = session.query(Project).get(self.project.id)
        if project is None:
            self.close()
            return

        store = self.bugtasks_store
        store.clear()

        bugtask_set = project.bugtask_set
        self.bugtasks_label.set_text('{0} bug tasks found'
                                      .format(bugtask_set.count()))

        for bugtask in bugtask_set:
            bug = bugtask.bug
            bugview = bug.bugview_set.first()
            if bugview is None:
                tags = ''
            else:
                bugviewtag_set = (session.query(Tag).join(BugViewTag)
                                  .filter(BugViewTag.bugview == bugview)
                                  .order_by(Tag.name).all())
                tags = ', '.join([bugviewtag.name
                                  for bugviewtag in bugviewtag_set])
            bug_id = bugtask.bug_id
            bugtaskview = bugtask.bugtaskview_set.first()
            status = bugtaskview.status.name
            importance = bugtaskview.importance.name
            if bugtaskview.assignee:
                assignee = bugtaskview.assignee.name
            else:
                assignee = ''
            if bugtaskview.milestone:
                milestone = bugtaskview.milestone.name
            else:
                milestone = ''
            store.append(None, (bug_id, status, importance,
                                assignee, milestone, tags))

    def plot_cb(self, widget):
        """
        Open a new tab to display bug statistics about this project
        """
        notebook = self.parent_notebook
        tab_id = PlotTab.get_id(self.project.name)
        notebook.open_tab(tab_id,
                          PlotTab, (self.project, self.application,))
