"""
Notebook tabs classes
"""
import os
from gi.repository import Gtk

from ...uifile import UIFile, Image
from ...db.model import Project, Milestone, Person, Bug


class Tab(UIFile):
    """
    Notebook tab with a close button

    Tab shows the contents of a TreeView row
    """
    def __init__(self, ui_basename, icon_image, application):
        """
        Initialize tab object
        """
        ui_filename = os.path.join('tab', ui_basename)
        UIFile.__init__(self, ui_filename)
        self._init_header(icon_image)
        self.parent_notebook = application.right_notebook
        self.application = application

    def _init_header(self, icon_image):
        """
        Create tab header with a close button
        """
        header = Gtk.HBox()
        header.set_spacing(2)
        title_label = Gtk.Label()
        button_image = Gtk.Image()
        button_image.set_from_stock(Gtk.STOCK_CLOSE, Gtk.IconSize.MENU)
        close_button = Gtk.Button()
        close_button.set_image(button_image)
        close_button.set_relief(Gtk.ReliefStyle.NONE)
        self.connect(close_button, 'clicked', self.close_cb)

        header.pack_start(icon_image,
                          expand=True, fill=True, padding=0)
        header.pack_start(title_label,
                          expand=True, fill=True, padding=0)
        header.pack_end(close_button,
                        expand=False, fill=False, padding=0)
        header.show_all()

        self.header = header
        self.title_label = title_label
        self.close_button = close_button

    def _get_icon_image(self, icon_basename):
        """
        Return tab icon image based on filename
        """
        return Image(icon_basename)

    @property
    def title(self):
        """
        Get tab title
        """
        return self.title_label.get_text()

    @title.setter
    def title(self, new_title):
        """
        Set tab title
        """
        self.title_label.set_text(new_title)

    def close(self):
        """
        Close tab
        """
        self.parent_notebook.close_tab(self)

    def close_cb(self, button):
        """
        Close tab when close button is clicked
        """
        self.close()

    def destroy(self):
        """
        Disconnect all signal handlers and destroy widgets

        Once the tab is destroyed, an attempt call them will still be
        made by ``Gtk`` when their associated signals are emitted. To
        avoid this, the handlers are cleanly disconnected.
        """
        self.disconnect_all()
        self.page.destroy()
        self.header.destroy()

    def row_activated_cb(self, treeview, path, view_column):
        """
        Open a notebook tab display detailed information
        regarding the item that has been activated
        """
        store = treeview.get_model()
        row = store[path]

        session = self.application.db.session
        column_title = view_column.get_title()
        if column_title == 'Bug ID':
            from .bug import BugTab
            tab_cls = BugTab
            db_row = session.query(Bug).get(row.bug_id)

        elif column_title == 'Project':
            from .project import ProjectTab
            tab_cls = ProjectTab
            db_row = session.query(Project).filter_by(name=row.project).one()

        elif column_title == 'Assignee' and row.assignee:
            from .person import PersonTab
            tab_cls = PersonTab
            db_row = session.query(Person).filter_by(name=row.assignee).one()

        elif column_title == 'Milestone' and row.milestone:
            from .milestone import MilestoneTab
            tab_cls = MilestoneTab
            if isinstance(self.db_row, Project):
                # Project tab: project is already stored in the tab
                project = self.db_row
            else:
                # Person tab: project name is in the store row
                # Bug tab: project name in the store row is empty.
                # It's in the parent row instead
                project_name = row.project or row.parent.project
                project = (session.query(Project)
                           .filter_by(name=project_name).one())
            db_row = (session.query(Milestone)
                      .filter_by(name=row.milestone, project=project).one())
        else:
            db_row = None

        if db_row is not None:
            notebook = self.parent_notebook
            tab_id = DBTab.get_id(db_row)
            notebook.open_tab(tab_id, tab_cls, (db_row, self.application))


class DBTab(Tab):
    """
    Notebook tab that displays information from a record in the database
    """
    def __init__(self, db_row, application):
        """
        Initialize tab object
        """
        # UI files are named after the db table they rely on
        ui_basename = '{0}.ui'.format(db_row.__tablename__)
        icon_image = self._get_icon_image(db_row)
        super(DBTab, self).__init__(ui_basename, icon_image, application)

        self.db_row = db_row
        self.id = self.get_id(db_row)

    def _get_icon_image(self, db_row):
        """
        Return tab icon image
        """
        return Image(self.ICON_BASENAME)

    @staticmethod
    def get_id(db_row):
        """
        Return tab id based on the database table and row id
        """
        return '{0}-{1}'.format(db_row.__tablename__,
                                db_row.id)
