"""
``Gtk.TreeStore`` specialized classes
"""
from gi.repository import Gtk

from .row import TreeModelRowWrapper


class TreeStore(Gtk.TreeStore):
    """
    ``Gtk.TreeStore`` wrapper

    It provides some additional features that makes it easier to use:

    * Keeps columns metadata so that column indexes can be accessed by name
    * Breadth first and depth first iteration. No need to recursively
      iterate through each row children anymore
    * Calculation of number of rows for all the tree (not just the
      number of rows in the top level of the tree)
    """

    def __init__(self, name, columns_metadata):
        """
        Initialize TreeStore object
        """
        column_types = [column.type for column in columns_metadata]
        Gtk.TreeStore.__init__(self, *column_types)

        self.set_name(name)
        column_names = [column.name for column in columns_metadata]
        if len(column_names) != len(set(column_names)):
            raise ValueError('All columns must have a different name: {0}'
                             .format(column_names))

        column_index = dict((name, index)
                            for index, name in enumerate(column_names))

        # Extract key column
        key_column = None
        key_columns = [column.name
                       for column in columns_metadata
                       if column.key]
        if len(key_columns) > 1:
            raise ValueError("Multiple key columns aren't allowed: {0}"
                             .format(', '.join(key_columns)))
        if key_columns:
            key_column = key_columns[0]

        # Set key column index
        if key_column is None:
            key_column_index = None
        elif isinstance(key_column, int):
            key_column_index = key_column
        elif isinstance(key_column, str):
            key_column_index = column_index[key_column]
        else:
            raise ValueError('Unexpected key column value: {0}'
                             .format(key_column))

        self.columns_metadata = columns_metadata
        self.column_index = column_index
        self.column_names = column_names
        self.key_column_index = key_column_index
        self.index = {}

    def __getitem__(self, key):
        """
        Get row in the store by key
        """
        return TreeModelRowWrapper(Gtk.TreeModel.__getitem__(self, key))

    def __getattr__(self, name):
        """
        Get column indexes by name
        """
        if name not in self.column_index:
            raise AttributeError('Unexpected column name: {0}'.format(name))

        return self.column_index[name]

    def __iter__(self):
        """
        Deletegate iteration to Gtk.TreeStore
        """
        return (TreeModelRowWrapper(row)
                for row in Gtk.TreeStore.__iter__(self))

    @property
    def n_rows(self):
        """
        Get the number of rows in the store (including all children
        rows).

        Note: ``store.n_rows`` != ``len(store)`` if the tree structure
        is not flat.
        """
        return len(list(self.bfiter()))

    def bfiter(self, root=None):
        """
        Breadth first iteration
        """
        if root is None:
            root = self

        queue = []
        for row in root:
            yield row
            queue.extend(list(row.iterchildren()))

        if queue:
            for row in self.bfiter(queue):
                yield row

    def dfiter(self, root=None):
        """
        Depth first iteration
        """
        if root is None:
            root = self

        for row in root:
            yield row
            for child_row in self.dfiter(row.iterchildren()):
                yield child_row

    def _validate_columns(self, parent, row_tuple):
        """
        Validate column values for a given row

        The following checks are performed:
        * Value isn't already present in any key column
          for any child row of ``parent``
        * Value isn't empty for required columns
        """
        self._check_duplicated_values(parent, row_tuple)
        self._check_required_columns(row_tuple)

    def _check_required_columns(self, row_tuple):
        """
        Check that all required columns have non-empty values
        """
        for value, column_metadata in zip(row_tuple, self.columns_metadata):
            required = column_metadata.required
            if required and not value:
                raise ValueError('Value required for column {0!r}: {1!r}'
                                 .format(column_metadata.name, value))

    def _check_duplicated_values(self, parent, row_tuple):
        """
        Make sure that value isn't already present in any key column
        for any child row of ``parent``
        """
        # Check is only made is key column index is defined
        if self.key_column_index is None:
            return

        if parent is None:
            rows = self
        else:
            rows = self[parent].iterchildren()

        row_tuple_key_column = row_tuple[self.key_column_index]
        for row in rows:
            row_key_column = row[self.key_column_index]
            if row_key_column == row_tuple_key_column:
                raise ValueError('Duplicated key column value detected: {0}'
                                 .format(row_key_column))

    def append(self, parent, row_tuple):
        """
        Append row and emit and update index if there's a key row
        """
        self._validate_columns(parent, row_tuple)
        row_iter = Gtk.TreeStore.append(self, parent, row_tuple)

        if (self.key_column_index is not None
            and parent is None):
            row = self[row_iter]
            key = row[self.key_column_index]
            if key is not None:
                self.index[key] = row.reference

        return row_iter

    def remove(self, iterator):
        """
        Update index and remove row
        """
        row = self[iterator]

        if self.key_column_index is not None:
            key = row[self.key_column_index]
            assert key in self.index, 'Row not found: {0}'.format(key)
            assert row.reference == self.index[key], \
                ('Index integrity check failure: {0}, {1}'
                 .format(row.reference.path, self.index[key].path))
            del self.index[key]

        Gtk.TreeStore.remove(self, iterator)

    def set(self, iterator, *data):
        """
        Set the value of one or more row columns
        """
        assert len(data) % 2 == 0, \
            'Unexpected data length: {0}'.format(data)

        pairs = zip(data[0::2], data[1::2])

        # Check that required columns aren't empty
        # (it's assumed that empty values are the ones that evaluate to false)
        for column_index, value in pairs:
            required = self.columns_metadata[column_index].required
            if (required and not value):
                raise ValueError, \
                    ('Value required for column {0!r}: {1!r}'
                     .format(self.column_names[column_index], value))

        Gtk.TreeStore.set(self, iterator, *data)

    def set_value(self, iterator, column_index, value):
        """
        Set value of one column of a row
        """
        # Check that required columns aren't empty
        # (it's assumed that empty values are the ones that evaluate to false)
        required = self.columns_metadata[column_index].required
        if (required and not value):
            raise ValueError, \
                ('Value required for column {0!r}: {1!r}'
                 .format(self.column_names[column_index], value))

        Gtk.TreeStore.set_value(self, iterator, column_index, value)


class TreeStoreColumnMetadata(object):
    """
    TreeStoreColumnMetadata objects are used to describe TreeStore columns.

    The following parameters are used:

    * name: Column name. It's used in the store to access to column
      index by name.
    * type: Column type. All elements in the column must be instances
      of the same type or any of its subclasses.
    * key (optional): Defines a key column, that is, a column for
      which duplicated values aren't allowed at the same level of the
      tree.
    * required (optional): Defines a column that always requires a
      non-empty value. All the values that evaluate to ``False`` are
      the ones considered empty.
    """
    def __init__(self, name, type, key=False, required=False):
        self.name = name
        self.type = type
        self.key = key

        # A  key row is assumed to required, that is, not empty
        self.required = required or key
