"""
Tree model row-related classes
"""
from gi.repository import Gtk


class TreeModelRowWrapper(object):
    """
    ``Gtk.TreeModelRow`` wrapper with some additional funcionality:

     * Rows are compared by their contents
     * Attribute access using column name (as defined in `TreeStore` object)
    """
    def __init__(self, row):
        """
        Initialize TreeModelRowWrapper object
        """
        object.__setattr__(self, '_row', row)

    def __eq__(self, other):
        """
        Compare TreeModelRow objects
        Rows are considered to be equal if their contents are equal
        """
        return (self.model == other.model
                and self.path == other.path)

    def __getattr__(self, name):
        """
        Delegate attribute access to ``Gtk.TreeModelRow`` object or `TreeStore`
        """
        # Access column by name
        # using TreeStore attribute access
        store = self._row.model
        index = getattr(store, name, None)
        if index is not None:
            return self[index]

        return getattr(self._row, name)

    def __setattr__(self, name, value):
        """
        Delegate attribute assignment to `TreeStore`
        """
        # Access column by name
        # using TreeStore attribute access
        row = self._row
        store = row.model
        index = getattr(store, name)

        store.set_value(self.iter, index, value)

    def __getitem__(self, index):
        """
        Delegate index access to ``Gtk.TreeModelRow`` object
        """
        # Map column name to column index
        if isinstance(index, str):
            store = self._row.model
            index = getattr(store, index)
        return self._row[index]

    def __setitem__(self, index, value):
        """
        Delegate index assignment to `TreeStore` object
        """
        # Map column name to column index
        if isinstance(index, str):
            index = self.model.column_index[index]

        store = self.model
        store.set_value(self.iter, index, value)

    def __len__(self):
        """
        Delegate length to ``Gtk.TreeModelRow`` object
        """
        return self.model.get_n_columns()

    @property
    def next(self):
        """
        Get next row in the tree
        """
        next_row = self._row.next
        if next_row is not None:
            next_row = TreeModelRowWrapper(next_row)
        return next_row

    @property
    def parent(self):
        """
        Get parent row
        """
        parent = self._row.parent
        if parent is not None:
            parent = TreeModelRowWrapper(parent)
        return parent

    def iterchildren(self):
        """
        Iterate through row children
        """
        return (TreeModelRowWrapper(child_row)
                for child_row in self._row.iterchildren())

    def iterancestors(self):
        """
        Iterate through row ancestors
        """
        parent_row = self.parent
        while parent_row:
            yield parent_row
            parent_row = parent_row.parent

    def bfiter(self, root=None):
        """
        Breadth first iteration
        """
        for row in self.model.bfiter(self.iterchildren()):
            yield row

    def dfiter(self):
        """
        Depth first iteration
        """
        for row in self.model.dfiter(self.iterchildren()):
            yield row

    @property
    def reference(self):
        """
        Get a reference object for this row
        """
        return TreeRowReferenceWrapper(self.model, self.path)

    @property
    def n_children(self):
        """
        Get the number of children under this row (not recursive)
        """
        return len(list(self.iterchildren()))

    @property
    def n_rows(self):
        """
        Get the number of rows under this row (recursive)
        """
        return len(list(self.dfiter()))

    def remove(self):
        """
        Remove row from model

        This is just a shortcut to avoid using the following less readable
        code::

         row.model.remove(row.iter)
        """
        self.model.remove(self.iter)


class TreeRowReferenceWrapper(object):
    """
    A ``Gtk.TreeRowReference`` wrapper
    that compares to another one based on model and path
    """
    def __init__(self, model, path):
        """
        Initialize TreeRowReference object
        """
        self._reference = Gtk.TreeRowReference.new(model, path)

    def __eq__(self, other):
        """
        Compare two references by model and path
        """
        assert isinstance(other, TreeRowReferenceWrapper), \
            'Unexpected row reference: {0}'.format(other)
        return (self.model == other.model
                and self.path == other.path)

    def valid(self):
        """
        Check if the reference still points to a valid row
        """
        return self._reference.valid()

    @property
    def model(self):
        """
        Get model
        """
        return self._reference.get_model()

    @property
    def path(self):
        """
        Get path
        """
        return self._reference.get_path()

    @property
    def row(self):
        """
        Get row
        """
        store = self.model
        path = self.path
        return TreeModelRowWrapper(store[path])
